import pytest
import numpy as np
import types

from src.chaotic_map import ChaoticMap
from src.plotting.plotting import ChaoticMapPlot

def double_function(x, r):
    x, y = x
    r, w = r
    return r * x * (1 - x), w * np.sin(np.pi * y)

@pytest.fixture
def chaotic_map_plot():
    chaotic_map = ChaoticMap(double_function)
    return ChaoticMapPlot(chaotic_map)


def test_init_wrong_type_input(chaotic_map_plot):
    """ Check if using a function as input for the class raises the appropriate error
    """
    with pytest.raises(TypeError):
        ChaoticMapPlot(lambda x: x)

@pytest.mark.utility
def test__find_iterable_single_iterable(chaotic_map_plot):
    """Test that the iterable is found"""
    assert chaotic_map_plot._find_iterable_([(1, 2, 3)]) == (0, (1, 2, 3))


@pytest.mark.utility
def test__find_iterable__from_list(chaotic_map_plot):
    """Find the location and iterable when iterable is in a list"""
    assert chaotic_map_plot._find_iterable_([(1, 2), 3, 4]) == (0, (1, 2))
    assert chaotic_map_plot._find_iterable_([1, (3, 4), 5]) == (1, (3, 4))
    assert chaotic_map_plot._find_iterable_([1, 4, 6, (3, 4), 5]) == (3, (3, 4))


@pytest.mark.utility
def test__find_iterable__multiple_iterables(chaotic_map_plot):
    """Find the location and iterable when iterable is in a list"""
    assert chaotic_map_plot._find_iterable_([1, (3, 4, 6), (5, 6), 4]) == (1, (3, 4, 6))
    assert chaotic_map_plot._find_iterable_([1, 2, (5, 6), 7, (9, 10)]) == (2, (5, 6))


@pytest.mark.utility
def test__find_iterable_raises_error_no_iterable(chaotic_map_plot):
    """Test that the iterable is found"""
    with pytest.raises(TypeError):
        chaotic_map_plot._find_iterable_([1, 2, 3, 4])


@pytest.mark.utility
def test__find_iterable_raises_error_empty_input(chaotic_map_plot):
    """Test that the iterable is found"""
    with pytest.raises(TypeError):
        chaotic_map_plot._find_iterable_(tuple())


@pytest.mark.utility
def test__iterator_parameter_gen__returns_generator(chaotic_map_plot):
    """Test that the sequence is finite"""
    r = np.random.uniform(0, 3.99, [1, 20])
    sequence = chaotic_map_plot._iterator_parameter_gen_((r,))
    assert isinstance(sequence, types.GeneratorType)


@pytest.mark.utility
def test__iterator_parameter_gen_single_input(chaotic_map_plot):
    """Test that the sequence is finite"""
    r = (1, 2, 3, 4)
    sequence = chaotic_map_plot._iterator_parameter_gen_((r,))
    print(sequence)
    assert list(sequence) == [(1,), (2,), (3,), (4,)]


@pytest.mark.utility
def test__iterator_parameter_gen_iterator_with_floats(chaotic_map_plot):
    """Test that the sequence is finite"""
    r = (1, 2, 3, 4)
    sequence = chaotic_map_plot._iterator_parameter_gen_((2, r, 3))
    assert list(sequence) == [(2, 1, 3), (2, 2, 3), (2, 3, 3), (2, 4, 3)]


@pytest.mark.utility
def test__iterator_parameter_gen_multiple_iterators(chaotic_map_plot):
    """Test that the sequence is finite"""
    r = (1, 2, 3, 4)
    sequence = chaotic_map_plot._iterator_parameter_gen_((2, r, 3, (1, 2, 3)))
    assert list(sequence) == [
        (2, 1, 3, (1, 2, 3)),
        (2, 2, 3, (1, 2, 3)),
        (2, 3, 3, (1, 2, 3)),
        (2, 4, 3, (1, 2, 3)),
    ]



""" ...
"""

from typing import Union, Tuple, Iterable

import itertools
import numpy as np


class ChaoticMap:
    """Define, iterate and analyze the dynamics of discrete chaotic maps"""

    def __init__(self, iteration):
        """Define a chaotic map using the iteration function"""
        self.iteration = iteration

    def next_point(self, point: Tuple[float], pams: Tuple[float]):
        """Get the next trajectory point

        Args:
        point (tuple[float]): The initial trajectory point
        pams (tuple[float]): Parameters for the map

        Returns:
        The next trajectory point

        Raises:
        ValueError: If any point generated is NaN
        """
        next_point = self.iteration(point, pams)
        if np.any(np.isnan(next_point)):
            raise ValueError("Encountered NaN value in iteration")
        return next_point

    def trajectory(
        self, point: Tuple[float], pams: Tuple[float], num_point: int = 2000
    ):
        """Generate a finite sequence of trajectory points

        Args:
        point (tuple[float]): The initial trajectory point
        pams (tuple[float]): Parameters for the map
        num_point (int, optional): number of points to generate, Default=2000

        Returns:
        A generator of trajectory points

        Raises:
        ValueError: If any element of the sequence is NaN
        """

        for _ in range(num_point):
            if np.any(np.isnan(point)):
                raise ValueError("Encountered NaN in infinite sequence")
            yield point
            point = self.iteration(point, pams)

    def sequence_gen_infinite(self, point: Tuple[float], pams: Tuple[float]):
        """Generate an infinite sequence of trajectory points

        Args:
        point (tuple[float]): The initial trajectory point
        pams (tuple[float]): Parameters for the map

        Returns:
        A generator of trajectory points

        Raises:
        ValueError: If any point generated is NaN
        """
        while True:
            if np.any(np.isnan(point)):
                raise ValueError("Encountered NaN")
            yield point
            point = self.iteration(point, pams)

    
    @staticmethod
    def move_by_d(point: Tuple[float], displace: float) -> Tuple[float]:
        """Alter the coordinates of a tuple point to move it by eucledean distance d

        Args:
        point (tuple[float]): The point to move
        displace (float): The distance to move the point

        Returns:
        The moved point
        """
        new_point = [
            point[k] + (-1) ** k * displace / np.sqrt(len(point))
            for k in range(len(point))
        ]
        new_point = tuple(new_point)
        return new_point

    @staticmethod
    def tuple_dist(tuple_1: Tuple[float], tuple_2: Tuple[float]) -> float:
        """Calculate the eucledean distance between two tuples

        Args:
        tuple_1 (tuple[float]): The first tuple
        tuple_2 (tuple[float]): The second tuple

        Returns:
        (float) The eucledean distance between the two tuples
        """

        if not len(tuple_1) == len(tuple_2):
            raise ValueError("Tuples must be of the same length to define distance")
        array_1, array_2 = np.array(tuple_1), np.array(tuple_2)
        return np.linalg.norm(array_1 - array_2)

    @staticmethod
    def alter_tuple_value(point: Tuple[float], idx: int, displacement: float):
        """Change a value of a tuple element at a specific index"""
        if not point:
            raise ValueError("Empty tuple entered")
        
        new_point = [
            point[k] + displacement if k == idx else point[k] for k in range(len(point))
        ]
        new_point = tuple(new_point)
        return new_point

    def approximate_partial_derivative(
        self, point: Tuple[float], pams: Tuple[float], **kwargs
    ) -> float:
        """Approximate the partial derivative of the chaotic map at a point

        The method used here is the finite difference method.
        I actually have no idea what algorithm is implemented tho, need to check.

        Args:
        point (tuple[float]): The point to calculate the partial derivative at
        pams (tuple[float]): The parameters of the function
        which_num (int, optional): The index of the numerator, Default=0
        which_den (int, optional): The index of the denominator, Default=0
        step (float, optional): The step size, Default=1e-4

        Returns:
        The partial derivative of the function at the point
        """
        which_num = kwargs.get("which_num", 0)
        which_den = kwargs.get("which_den", 0)
        step = kwargs.get("step", 1e-4)

        moved_point_1 = ChaoticMap.alter_tuple_value(point, which_den, step)
        moved_point_2 = ChaoticMap.alter_tuple_value(point, which_den, -step)

        next_moved_point_1 = self.iteration(moved_point_1, pams)
        next_moved_point_2 = self.iteration(moved_point_2, pams)

        return (next_moved_point_1[which_num] - next_moved_point_2[which_num]) / (
            2 * step
        )

    def approximate_jacobian(
        self, point: Tuple[float], pam: Tuple[float], step: float = 1e-4
    ) -> np.array:
        """Approximate the Jacobian of the chaotic map at a point

        The method used here is the finite difference method.

        Args:
        point (tuple[float]): The point to calculate the Jacobian at
        pam (tuple[float]): The parameters of the function
        step (float, optional): The step size, Default=1e-4

        Returns:
        The Jacobian of the function at the point
        """
        ind = range(len(point))
        j = np.array(
            [
                self.approximate_partial_derivative(
                    point, pam, which_num=i, which_den=j, step=step
                )
                for i in ind
                for j in ind
            ]
        )
        j = j.reshape((len(point), len(point)))
        return j

    def approximate_lyapunov_exponents(
        self, point: Tuple[float], pams: Tuple[float], **kwargs
    ) -> np.array:
        """Approximate the Lyapunov exponents of the chaotic map at a point

        Args:
        point (tuple[float]): The point to calculate the Lyapunov exponents at
        pams (tuple[float]): The parameters of the function
        discard (int, optional): The number of iterations to discard, Default=100
        num_points (int, optional): The number of points to use in the approximation, Default=1000

        step (float, optional): The step size, Default=1e-4
        """

        num_points = kwargs.get("num_points", 1000)
        discard = kwargs.get("discard", 100)
        step = kwargs.get("step", 1e-04)

        gen = self.trajectory(point, pams, num_points)
        gen = itertools.islice(gen, discard, None)

        diagonal_elements = []

        q_diagonal = np.identity(len(point))

        for traj_poin in gen:
            jacobian = self.approximate_jacobian(traj_poin, pams, step)
            updated_q = jacobian @ q_diagonal
            q_diagonal, r_factor = np.linalg.qr(updated_q)
            diagonal_elements.append(np.diagonal(r_factor))

        les = np.zeros(diagonal_elements[0].shape)

        for diag in diagonal_elements:
            les += np.log(np.abs(diag))

        return les / num_points

    def is_divergent(
        self, point: Tuple[float], pams: Tuple[float], num_points: int = 2000
    ) -> bool:
        """Check if intial condition and parameter selections lead to divergent trajectory.

        Args:
        point (tuple[float]): The initial condition
        pams (tuple[float]): The parameters of the chaotic map
        num_points (int, optional): The number of points to check, Default=2000

        Returns:
        (bool) True if the trajectory is divergent, False otherwise
        """
        try:
            sequence = list(self.trajectory(point, pams, num_point=num_points))
        except ValueError:
            return True

        if -np.inf in min(sequence) or np.inf in max(sequence):
            return True
        return False



""" This module provides utilities for plotting and visualizing chaotic maps."""
from typing import Union, Tuple, Iterable

import warnings
import matplotlib.pyplot as plt

from typing import List, Tuple, TypeVar, Union
Figure = TypeVar("Figure")

from chaotic_maps.chaotic_maps import ChaoticMap


class ChaoticMapPlot:
    """This class takes a chaotic map object and provides the
    methods for creating plots of its dynamical behavior.
    """

    def __init__(self, chaotic_map: ChaoticMap):
        """Initialize the object for generating the plots

        Args:
            chaotic_map (ChaoticMap): The chaotic map object

        Raises:
            TypeError -- If the chaotic map is not of type ChaoticMap
        """
        if not isinstance(chaotic_map, ChaoticMap):
            raise TypeError("ChaoticMap object required as input")
        self.map = chaotic_map


    @staticmethod
    def _find_iterable_(
        pams: Tuple[Union[Iterable, float]]
    ) -> Union[Tuple[Iterable, int], None]:
        """Locate the first iteratble element of a tuple

        Args:
        pams (tuple[Union[Iterable, float]]): The tuple to search

        Returns:
        The first iterable and its index on the input tuple

        Raises:
        ValueError: If no iterable is found
        """
        loc, par_values = None, None
        for num, item in enumerate(pams):
            if hasattr(item, "__iter__"):
                loc, par_values = num, item
                return loc, par_values
        if par_values is None:
            raise TypeError("No iterator is given")


    @staticmethod
    def _iterator_parameter_gen_(pams: Tuple[Union[Iterable, float]]) -> Tuple[float]:
        """Generate every possible combination of tuples from a tuple containing an iterable

        Args:
        pams (tuple[Union[Iterable, float]]): The tuple to search

        Yields:
        The next parameter tuple
        """

        loc, par_values = ChaoticMapPlot._find_iterable_(pams)
        for par_value in par_values:
            yield tuple(par_value if i == loc else pams[i] for i in range(len(pams)))


    def plot_trajectory(
        self,
        init_point: Tuple[float],
        parameter: Tuple[float],
        num_points: int = 100,
        which_var: int = 0,
        fig: Union[Figure, None] = None,
        **kwargs
    ) -> Figure:
        """ Plot the trajectory of the selected system variable

        Args:
        init_point:
        The initial point of the trajectory

        parameter:
        The tuple representing the parameters for the cahotic map

        num_points (Optional):
        The number of trajectrory points to be plotted (Default 2000)

        which_var (Optional):
        The variable to be plotted (Default 0)

        fig (Optional):
        The figure in which the scatter plot is to be created.

        kwags:
        Named arguments to alter the plot appearance, refer to matplotlib
        scatter plot for options.


        Returns:

        fig (Figure):
        The figure object that contains the return map
        """
        seq_gen = self.map.trajectory(init_point, parameter, num_points)

        points = [point[which_var] for point in seq_gen]

        if fig is None:
            fig = plt.figure()
        else:
            plt.figure(fig.number)

        plt.plot(points, **kwargs)

        return fig

    def bifurcation_dict(
        self,
        start_point: Tuple[float],
        parameters: Tuple[Union[float, Iterable]],
        num_points: int = 500,
        points_to_skip: int = 50,
    ):
        """Get a dictionary with parameter values as key and trajectories.

        Args:
            start_point (Tuple[float]): The starting point of the trajectory
            parameters (Tuple[Union[float, Iterable]]): The parameters to vary
            num_points (int, optional): The number of points for each trajectory. Defaults to 500.
            points_to_skip (int, optional): Number of points to discard. Defaults to 50.

        Returns:
            dict: A dictionary with parameter values as keys and generators of
            trajectory objects as values

        Raises:
            ValueError: If the number of points to be discarded are more that
            the number of points to generate
        """
        if points_to_skip >= num_points:
            raise ValueError(
                "The number of total points generated has to be greater than"
                + "the number of points skipped"
            )

        bif_dict = {}
        pam_loc, _ = self._find_iterable_(parameters)
        for pam in self._iterator_parameter_gen_(parameters):
            try:
                point_gen = self.map.trajectory(start_point, pam, num_points)
                for _ in range(points_to_skip):
                    next(point_gen)
                bif_dict[pam[pam_loc]] = point_gen
            except Exception as exc:  # If an exception is raised skip this parameter
                warnings.warn(f"Skipping parameter {pam} due to {exc}")
                continue
        return bif_dict

    def bifurcation_diagram(
        self, start_point: Tuple[float], pams: Tuple[Union[float, Iterable]], **kwargs
    ):
        """Create the bifurcation diagram for the chaotic map on a
        given set of parameters.

        Args:
            start_point (Tuple[float]): The starting point of the trajectory
            pams (Tuple[Union[float, Iterable]]): The parameters to vary
            **kwargs: Additional keyword arguments for the plot

        Returns:
            matplotlib.pyplot.Figure: The figure object
        """

        which_var = kwargs.pop("which_var", 0)

        num_points = kwargs.pop("num_points", 500)

        points_to_skip = kwargs.pop("points_to_skip", 50)

        fig = kwargs.pop("fig", None)

        # Find the position of the iterable in the pams tuple
        loc, _ = self._find_iterable_(pams)

        # Create the bifurcation dictionary, where the keys are the parameter tuples
        bif_dict = self.bifurcation_dict(start_point, pams, num_points, points_to_skip)

        if fig is None:
            fig = plt.figure()
        for key, value in bif_dict.items():
            pam = key
            points = [val[which_var] for val in value]
            ell = len(points)
            plt.scatter(ell * [pam], points, **kwargs)
        return fig

    def return_map(
        self,
        init_point: Tuple[float],
        parameter: Tuple[float],
        num_points: int =2000,
        which_var: int = 0,
        fig: Figure = None,
        **kwargs
    ) -> Figure:
        """Plot the return map diagram of the given chaotic map on a given or new figure

        The return map is a diagram in which the x axis represent a the value at k-th step
        and the y axis is the value of the trajectory at the (k+1)-th step.
        The figure is thus a scatter plots where (x, y) are subsequent trajectory points.


        Args:

        init_point:
        The initial point of the trajectory

        parameter:
        The tuple representing the parameters for the cahotic map

        num_points (Optional):
        The number of trajectrory points to be plotted (Default 2000)

        which_var (Optional):
        The variable to be plotted (Default 0)

        fig (Optional):
        The figure in which the scatter plot is to be created.

        kwags:
        Named arguments to alter the plot appearance, refer to matplotlib
        scatter plot for options.


        Returns:

        fig (Figure):
        The figure object that contains the return map
        """
        seq_gen = self.map.trajectory(init_point, parameter, num_points)
        points1 = [point[which_var] for point in seq_gen]
        points2 = points1[1:]
        points1 = points1[:-1]
        if fig is None:
            fig = plt.figure()
        else:
            plt.figure(fig.number)
        plt.scatter(points1, points2, **kwargs)
        return fig

    @staticmethod
    def _cobweb_line_(xs: float, ys: float, zs: float
    ) -> Tuple[List[float], List[float]]:
        """Create a line for the cobweb plot

        Given three consecutive trajectory points for one variable, create the
        lists that define the trajectory line for the cobweb diagram.
        """

        x_locs = [xs, ys, ys]
        y_locs = [ys, ys, zs]

        return x_locs, y_locs

    def cobweb_diagram(
        self,
        init_point: Tuple[float],
        parameter: Tuple[float],
        num_points: int = 50,
        which_var: int = 0,
        fig: Union[Figure, None] = None,
        **kwargs
    ) -> Figure:
        """ Create the cobweb diagram for the selected initial condition and parameter
        values of the map.

        For details about the Cobweb diagram, refer to

        Args:
        init_point:
        The initial point of the trajectory

        parameter:
        The tuple representing the parameters for the cahotic map

        num_points (Optional):
        The number of trajectrory points to be plotted (Default 2000)

        which_var (Optional):
        The variable to be plotted (Default 0)

        fig (Optional):
        The figure in which the scatter plot is to be created.

        kwags:
        Named arguments to alter the plot appearance, refer to matplotlib
        scatter plot for options.


        Returns:

        fig (Figure):
        The figure object that contains the return map
        """
        seq_gen = self.map.trajectory(init_point, parameter, num_points)

        points = []
        for k in range(num_points):
            point = next(seq_gen)
            points.append(point[which_var])
        # points = [point[which_var] for point in seq_gen]
        x_plot = []
        y_plot = []
        for k in range(len(points) - 2):

            xs, ys, zs = points[k : k + 3]
            x_n, y_n = self._cobweb_line_(xs, ys, zs)
            x_plot.extend(x_n)
            y_plot.extend(y_n)

        if fig is None:
            fig = plt.figure()
        else:
            plt.figure(fig.number)

        plt.plot(x_plot, y_plot, **kwargs)
        plt.xlim(min(x_plot), max(x_plot))
        plt.ylim(min(y_plot), max(y_plot))
        plt.plot([min(x_plot), max(x_plot)], [min(x_plot), max(x_plot)], "k")

        return fig

    def lyapunov_exponent_dict(
        self,
        start_point: Tuple[float],
        parameters: Tuple[Union[Tuple, float]],
        num_points: int = 500,
        points_to_skip: int = 50,
    ) -> dict:
        """Get a dictionary with parameter values as key and the lyapunov exponent as value.

        Args:
            start_point (Tuple[float]): The starting point of the trajectory
            parameters (Tuple[Union[Tuple, float]]): The parameters to vary
            num_points (int, optional): The number of points for each trajectory. Defaults to 500.
            points_to_skip (int, optional): Number of points to discard. Defaults to 50.

        Returns:
            dict: A dictionary with parameter values as keys and generators of

        Raises:
            ValueError: If the number of points to be discarded are more that the number of points kept
        """
        if points_to_skip >= num_points:
            raise ValueError(
                "The number of total points generated has to be greater than the"
                + "number of points skipped"
            )
        le_dict = {}
        for pam in self._iterator_parameter_gen_(parameters):
            try:
                le_now = self.map.approximate_lyapunov_exponents(
                    start_point, pam, num_points=1000, discard=100, h=1e-04
                )
                pam_loc, _ = self._find_iterable_(parameters)
                le_dict[pam[pam_loc]] = le_now
            except Exception as exc:
                warnings.warn(f"Skipping parameter {pam} due to {exc}")
                continue
        return le_dict

    def lyapunov_exponent_plot(
        self,
        start_point: Tuple[float],
        parameters: Tuple[Union[Tuple, float]],
        num_points: int = 500,
        points_to_skip: int = 50,
        fig: Union[Figure, None] = None,
        ** kwargs
        ) -> Figure:
        """Get a dictionary with parameter values as key and the lyapunov exponent as value.

        Args:
            start_point (Tuple[float]): The starting point of the trajectory
            parameters (Tuple[Union[Tuple, float]]): The parameters to vary
            num_points (int, optional): The number of points for each trajectory. Defaults to 500.
            points_to_skip (int, optional): Number of points to discard. Defaults to 50.
            fig (Figure, optional): The figure in which the Lyapunov exponent is plotted
            kwargs: plt.plot arguments
        Returns:
            A plt figure object containing the plot of the lyapunov exponent for the selected parameter

        """
        le_dict = self.lyapunov_exponent_dict(
            start_point,
            parameters,
            num_points,
            points_to_skip,
            )

        if fig is None:
            fig = plt.figure()
        else:
            plt.figure(fig.number)

        plt.plot(le_dict.keys(), le_dict.values(), **kwargs)

        return fig
